@@ -1,5 +1,5 @@
 package Data::Babel;
-our $VERSION='1.11';
+our $VERSION='1.12';
 $VERSION=eval $VERSION;         # I think this is the accepted idiom..
 #################################################################################
 #
@@ -20,24 +20,42 @@ use strict;
 use Class::AutoClass;
 use Carp;
 use Graph::Undirected;
-use List::MoreUtils qw(uniq);
+use Graph::Directed;
+use List::MoreUtils qw(uniq none);
 use List::Util qw(min);
 use Hash::AutoHash::Args;
-use Hash::AutoHash::MultiValued;
 use Data::Babel::Config;
 use Data::Babel::IdType;
 use Data::Babel::Master;
 use Data::Babel::MapTable;
 use Data::Babel::HAH_MultiValued;
+use Data::Babel::PrefixMatcher::Exact;
+use Data::Babel::PrefixMatcher::Trie;
 
 use vars qw(@AUTO_ATTRIBUTES @OTHER_ATTRIBUTES %SYNONYMS %DEFAULTS %AUTODB);
 use base qw(Data::Babel::Base);
 
 # name, id, autodb, verbose - methods defined in Base
-@AUTO_ATTRIBUTES=qw();
-@OTHER_ATTRIBUTES=qw(idtypes masters maptables schema_graph);
+#
+# NG 13-07-18: pdup removal
+# attributes to control partial duplicate removal algorithm
+#   pdups_group_cutoff - do not group results unless # rows > this
+#     set to 0 to always do grouping
+#   pdups_prefixmatcher_cutoff - run brute force algorithm unless # rows > this
+#     # rows is per group if grouping in effect
+#     set to 0 to always use prefix matching
+#   pdups_prefixmatcher_class - PrefixMatcher subclass used when prefix matching in effect
+#     final component of name. full name is Data::Babel::PrefixMatcher::<this>
+#     current choices: Trie (default), BinarySearchTree, BinarySearchList, PrefixHash
+# note that keep_pdups is passed argument to translate - not Babel attribute
+# defaults determined through (very limited) testing. hopefully not too far off...
+@AUTO_ATTRIBUTES=qw(pdups_group_cutoff pdups_prefixmatcher_cutoff );
+
+@OTHER_ATTRIBUTES=qw(idtypes masters maptables schema_graph pdups_prefixmatcher_class);
 %SYNONYMS=();
-%DEFAULTS=(idtypes=>[],masters=>[],maptables=>[],);
+%DEFAULTS=
+  (idtypes=>[],masters=>[],maptables=>[],
+   pdups_group_cutoff=>50,pdups_prefixmatcher_cutoff=>250,pdups_prefixmatcher_class=>'Trie');
 %AUTODB=(-collection=>'Babel',-keys=>qq(name string),-transients=>qq());
 Class::AutoClass::declare;
 
@@ -131,18 +149,26 @@ sub translate {
   # NG 12-08-22: okay to omit input_ids; same as input_ids_all=>1
   # confess "At least one of input_ids or input_ids_all must be set" if $ids_args==0;
   confess "At most one of input_ids or input_ids_all may be set" if $ids_args>1;
-  my $sql=$self->generate_query($args);
+
+  # NG 13-06-19: added $query_paths,$input,@outputs. needed for pdup elimination
+  # NG 13-06-21: replace $input_name,@output_names by $columns - SQL column names
+  #              of results table
+  my($sql,$query_paths,$columns)=$self->generate_query($args);
   my $dbh=$args->dbh || $self->autodb->dbh;
   # my $results=$args->count? $dbh->selectrow_array($sql): $dbh->selectall_arrayref($sql);
   # NG 13-06-11: added this code to display query
-  if ($self->verbose) {
-    print "$sql\n";
-  }
+  print "$sql\n" if $self->verbose;
   my $results=$dbh->selectall_arrayref($sql);
   confess "Database query failed:\n$sql\n".$dbh->errstr if $dbh->err;
-  unless ($args->validate) {
-    $results=@$results? $results->[0][0]: 0 if $args->count;
-  } else { 
+  $results=$self->remove_pdups($results,$query_paths,$columns,$args) unless $args->keep_pdups;
+
+  # NG 12-11-23: if validating, have to do count in code...
+  # NG 13-06-20: with pdup removal, have to do count in code always. 
+  #              kinda makes count pointless....
+  if ($args->count && $args->keep_pdups && !$args->validate) {
+    return @$results? $results->[0][0]: 0;
+  }
+  if ($args->validate) { 
     # cases:
     # 1) no filters. $results contains all valid input_ids. missing ids are invalid
     # 2) filters. $results contains subset of valid input_ids. have to query db to
@@ -177,20 +203,21 @@ sub translate {
       %id2valid=map {$_=>1} @$valid_ids;
       @input_ids=@$valid_ids unless defined $input_ids;
     }
+    # NG 13-06-21: moved addition of VALID column from generate_query to here
+    map {splice(@$_,1,0,1)} @$results;
     # NG 13-06-10: convert to lower case for case insensitive comparisons
     @missing_ids=grep {!$have_id{lc $_}} @input_ids;
     push(@$results,map {[$_,$id2valid{lc $_}||0,(undef)x$num_outputs]} @missing_ids);
- 
-    # post-processing
-    # 1) count. return number of rows
-    # 2) limit. select subset of rows
-    # 3) count+limit: return min of count, limit
-    if ($args->count) {
-      $results=scalar @$results;
-      $results=min($args->limit,$results) if defined $args->limit;
-    } elsif (defined $args->limit) {
-      @$results=@$results[0..min($args->limit,scalar @$results)-1];
-    }
+  }
+  # post-processing
+  # 1) count. return number of rows
+  # 2) limit. select subset of rows
+  # 3) count+limit: return min of count, limit
+  if ($args->count) {
+    $results=scalar @$results;
+    $results=min($args->limit,$results) if defined $args->limit;
+  } elsif (defined $args->limit) {
+    @$results=@$results[0..min($args->limit,scalar @$results)-1];
   }
   $results;
 }
@@ -235,9 +262,11 @@ sub generate_query {
   my $dbh=$self->autodb->dbh;
   my @columns=($input_name,@output_names);
   # splice in VALID column if validating
-  splice(@columns,1,0,'1 AS VALID') if $args->validate;
+  # NG 13-06-21: move addition of VALID column to post-processing
+  # splice(@columns,1,0,'1 AS VALID') if $args->validate;
   # MySQL doesn't allow duplicate columns in inner queries. sigh..
   @columns=uniq(@columns) if $args->count;
+  my $columns=\@columns;
   my $columns_sql=join(', ',@columns);
   
   my $input_master=$input_idtype->master;
@@ -247,8 +276,20 @@ sub generate_query {
     ($input_master->explicit || $input_idtype->degree>1 || @idtypes==1 || !defined($input_ids))?
       $input_master->tablename: ();
   # get rest of query by exploring query graph
-  push(@join_tables,map {$self->id2name($_)} 
-       $self->traverse($self->make_query_graph(@idtypes))) if @idtypes>1;
+  # NG 13:06-19: change one-liner to 'if' block as prelude to pdups removal
+  # push(@join_tables,map {$self->id2name($_)} 
+  #      $self->traverse($self->make_query_graph(@idtypes))) if @idtypes>1;
+  my($root,$query_graph,$query_paths);
+  if (@idtypes>1) {
+    # NOTE: input must be 1st idtype!
+    ($root,$query_graph)=$self->make_query_graph(@idtypes);
+
+    # NG 13-06-19: added make_query_paths. needed for pdup elimination
+    $query_paths=$self->make_query_paths($query_graph,$input_idtype,@output_idtypes)
+      unless $args->keep_pdups;
+
+    push(@join_tables,map {$self->id2name($_)} $self->traverse($root,$query_graph));
+  }
   # add in filter masters for any with histories
   push(@join_tables,map {$_->tablename} grep {$input_idtype != $_} @filter_idtypes);
   my $join_sql=join(' NATURAL LEFT OUTER JOIN ',@join_tables);
@@ -303,9 +344,16 @@ sub generate_query {
   # Note: 'AS T' needed at end of query because mySQL sees inner select as defining derived
   #       table and requires every derived table have alias
   # NG 12-11-23: if validating, have to do count in code...
-  $sql=qq(SELECT COUNT(*) FROM ($sql) AS T) if $args->count && !$args->validate;
+  # NG 13-06-20: with pdup removal, have to do count in code. kinda makes count pointless....
+  $sql=qq(SELECT COUNT(*) FROM ($sql) AS T)
+    if $args->count && $args->keep_pdups && !$args->validate;
   
-  $sql;
+  # NG 13-06-19: added $query_paths,$input_name,@output_names to return. 
+  #              needed for pdup elimination
+  # NG 13-06-21: replace $input_name,@output_names by $columns - SQL column names
+  #              of results table
+  # $sql;
+  ($sql,$query_paths,$columns);
 }
 # input is IdType object. handles history
 sub _generate_colname {
@@ -331,7 +379,7 @@ sub schema_graph {
 # query graph is a steiner minimum tree whose terminals are the input and output IdTypes
 # trivial to compute for __non_redundant__ schema. just prune back non-terminal leaves
 # first arg is input_idtype; rest are output_idtypes. specified as objects!
-# returns ($root,$query_graph)
+# returns ($root,$query_graph). note that $root needed by traverse
 sub make_query_graph {
   my $self=shift;
   my @terminals=map {$_->id} @_; # nodes. not objects
@@ -353,19 +401,190 @@ sub make_query_graph {
   #              to achieve correct UR semantics. processing moved to generate_query.
   #              now use any neighbor of input IdType
   my($root)=$query_graph->neighbors($input_node);
-#  # root processing: if input_idtype touches >1 maptable, add master as root
-#  #                  else use the input_idtype's maptable as root
-#   my $root;
-#   if ($query_graph->degree($input_node)>1) {
-#     my $master_obj=$input_idtype->master or 
-#       confess "No Master found for input IdType ".$input_idtype->name;
-#     $root=$master_obj->id;
-#     $query_graph->add_edge($root,$input_node);
-#   } else {
-#     ($root)=$query_graph->neighbors($input_node);
-#   } 
   ($root,$query_graph);
 }
+# NG 13-06-19: added make_query_paths 
+# enumerate root-to-leaf paths in results graph, extracted from query graph
+#   results graph is directed tree. nodes are idtypes. edges represent paths in query graph
+#   rooted at input
+# input, outputs are idtypes
+sub make_query_paths {
+  my $self=shift;
+  my($query_graph,$input_idtype,@output_idtypes)=@_;
+  my($root,$results_graph)=$self->make_results_graph($query_graph,$input_idtype,@output_idtypes);
+  my @paths=paths($root,$results_graph);
+  wantarray? @paths: \@paths
+}
+sub make_results_graph {
+  my $self=shift;
+  my($query_graph,$input_idtype,@output_idtypes)=@_;
+  my($input,@outputs)=map {$_->id} ($input_idtype,@output_idtypes);
+  my $results_graph=new Graph::Directed;
+  map {$results_graph->add_vertex($_)} ($input,@outputs);
+  elide($input,$query_graph,$results_graph);
+  # if input has history and also appears in output add edge from history to idtype
+  my $root;
+  if ($input_idtype->history && grep {$input_idtype==$_} @output_idtypes) {
+    $root='_X_'.$input_idtype->name;
+    $results_graph->add_edge($root,$input);
+  } else {
+    $root=$input;
+  }
+  ($root,$results_graph);
+}
+# TODO: belongs is some Util
+# group a list by categories returned by sub
+# has to be declared before use, because of prototype
+sub group (&@) {
+  my($sub,@list)=@_;
+  my %groups;
+  for (@list) {
+    my $group=&$sub($_);
+    my $members=$groups{$group} || ($groups{$group}=[]);
+    push(@$members,$_);
+  }
+  wantarray? %groups: \%groups;
+}
+# collapse paths in query_graph to edges in results_graph
+my %seen;
+sub elide {
+  my($root)=@_;
+  %seen=($root=>$root);
+  _elide($root,@_);
+}
+sub _elide {
+  my($prev,$node,$query_graph,$results_graph)=@_;
+  my @maptables=map {$seen{$_}=$_} grep {!$seen{$_}} $query_graph->neighbors($node);
+  my @idtypes=map {$seen{$_}=$_} grep {!$seen{$_}} map {$query_graph->neighbors($_)} @maptables;
+  # output idtypes already in results graph
+  my %groups=group {$results_graph->has_vertex($_)? 'output': 'internal'} @idtypes;
+  my @outputs=@{$groups{output}||[]};
+  my @internals=@{$groups{internal}||[]};
+  map {$results_graph->add_edge($prev,$_)} @outputs;
+
+  # TODO: don't recurse over leafs
+  # each output roots another traversal
+  map {_elide($_,$_,$query_graph,$results_graph)} @outputs;
+  # each internal continues current traversal
+  map {_elide($prev,$_,$query_graph,$results_graph)} @internals;
+  $results_graph;
+}
+
+# enumerate root to leaf paths in results graph (actually, any directed tree)
+sub paths {
+  my($root,$tree)=@_;
+  my @successors=$tree->successors($root);
+  my @paths=map {paths($_,$tree)} @successors;
+  @paths=@paths? _paths($root,@paths): [$root];
+  # strip off 'idtype:' prefix. no longer needed (since all nodes are idtypes)
+  @paths=map {[map {s/^idtype://; $_} @$_]} @paths;
+  wantarray? @paths: \@paths;
+}
+# expand $root, list of paths
+sub _paths {
+  my $root=shift;
+  map {unshift(@$_,$root); $_;} @_;
+}
+
+########################################
+# NG 13-07-18: remove partial duplicates (pdups)
+sub remove_pdups {
+  my $self=shift;
+  my($input,$query_paths,$columns,$args)=@_;
+  # if $query_paths is undef or <= 2 columns, no pdups possible
+  return $input unless $query_paths || @$columns>2;
+  my($group_cutoff,$prefixmatch_cutoff,$matcher_class)=
+    $self->get(qw(pdups_group_cutoff pdups_prefixmatcher_cutoff pdups_prefixmatcher_class));
+  my $exact_class='Data::Babel::PrefixMatcher::Exact';
+  my @matchers;
+  my @paths=@$query_paths;
+  my %col2idx=val2idx(@$columns);
+  my @path_idxs=map {[@col2idx{@$_}]} @paths; # used to extract cols for each path
+  my $output=[];
+
+  my(%groups,$skip0);
+  if (@$input<=$group_cutoff) {
+    %groups=('all',$input);
+  } else {
+    %groups=group {$_->[0]} @$input;
+    map {shift @$_} @path_idxs; # don't need 0'th element any more
+    $skip0=1;
+  }
+  for my $group (values %groups) {
+    push(@$output,@$group),next unless @$group>1;
+    my @rows=sort {num_undefs($a) <=> num_undefs($b)} @$group;
+    if (@$group<=$prefixmatch_cutoff) {
+      # brute force
+      my %rows=map {$_=>$rows[$_]} (0..$#rows);
+	for (my $i=0; $i<@rows-1; $i++) {
+	  next unless $rows{$i};
+	  for (my $j=$i+1; $j<@rows; $j++) {
+	    next unless $rows{$j};
+	    delete $rows{$j} if pdup($rows[$i],$rows[$j],$skip0);
+	  }}
+      push(@$output,values %rows);
+    } else {
+      # use prefix matching
+      if (@matchers) {
+       map {$_->reset} @matchers;
+     } else {
+       # optimize trivial matchers - if path of length 1, use exact hash 
+       @matchers=map {@$_==1? new $exact_class: new $matcher_class} @path_idxs;
+     }
+      my $row_idx=0;
+      for my $row (@rows) {
+	my @subrows=map {my @idxs=@{$_}; [@$row[@idxs]]} @path_idxs;
+	my $new=0;			# assume it's an old row - not pdup
+	my @subhits;
+	for (my $j=0; $j<@paths; $j++) {
+	  my $subrow=$subrows[$j];
+	  next if none {defined $_} @$subrow; # all NULL - wildcard - matches everything
+	  my $matcher=$matchers[$j];
+	  my $subhits=$matcher->get_data($subrow); # returns ARRAY of rowidxs
+	  push(@subhits,$subhits);
+	}
+	# compute intersection of subhits
+	my @hits=intersect(@subhits);
+	unless (@hits) {
+	  # it's a new row!
+	  push(@$output,$row);	# add to output
+	  # add to matchers
+	  for (my $j=0; $j<@paths; $j++) {
+	    my $subrow=$subrows[$j];
+	    my $matcher=$matchers[$j];
+	    $matcher->put_data($subrow,$row_idx);
+	  }}
+	$row_idx++;
+      }}} 
+  $output;
+}
+# calculate number of undefs in row
+sub num_undefs {scalar grep {!defined $_} @{$_[0]}}
+# row $j is partial-dup of $i if they agree wherever both defined, else $j is undef
+# $skip0 is 1 if $input grouped, else 0
+sub pdup {
+  my($rowi,$rowj,$skip0)=@_;
+  for(my $k=$skip0; $k<=@$rowi; $k++) {
+    return 0 if defined $rowj->[$k] && $rowi->[$k] ne $rowj->[$k];
+  }
+  1;
+}
+sub pdups_prefixmatcher_class {
+  my $self=shift;
+  if (@_) {
+    my $subclass=shift;
+    my $class='Data::Babel::PrefixMatcher::'.$subclass;
+    unless (eval "require $class") {
+      confess "BAD NEWS: PrefixMatcher subclass Trie not found: $@" if $subclass eq 'Trie';
+      carp "PrefixMatcher subclass $subclass not found. Falling back to Trie: $@";
+      $class='Data::Babel::PrefixMatcher::Trie';
+    }
+    $self->{pdups_prefixmatcher_class}=$class;
+  }
+  $self->{pdups_prefixmatcher_class};
+}
+
+########################################
 
 # make implicit masters
 #   every IdType needs a Master. if Master not defined explicitly, define it here
@@ -464,7 +683,7 @@ sub show_schema_graph {
 # can be called as function or method
 sub show_graph {
   my $graph=ref($_[0]) && $_[0]->isa('Graph')? $_[0]: $_[1];
-  print '  ',join("\n",map {_edge_str($graph,$_)} _sort_edges($graph->edges)),"\n";
+  print '  ',join("\n  ",map {_edge_str($graph,$_)} _sort_edges($graph->edges)),"\n";
 }
 sub _sort_edges {
   my @edges=map {$_->[0] le $_->[1]? $_: [$_->[1],$_->[0]]} @_;
@@ -578,8 +797,26 @@ sub _2idtype {
 #   $idtype->name;
 # }
 
-# TODO: belongs in some Util
+########################################
+# TODO: move these functions to some Util
 sub flatten {map {'ARRAY' eq ref $_? @$_: $_} @_;}
+# produce hash mapping each element of list to its position. doesn't worry about duplicates
+sub val2idx {
+  my $i=0;
+  my %val2idx=map {$_=>$i++} @_;
+  wantarray? %val2idx: \%val2idx;
+}
+# compute intersection of 2 or more lists
+# TODO: simplify interface with prototype
+sub intersect {
+  my %x=map {$_=>1} flatten(shift);
+  while (@_) {
+    my $y=shift;
+    %x=map {$_=>1} grep {$x{$_}} @$y;
+  }
+  my @x=keys %x;
+  wantarray? @x: \@x;
+}
 
 # NG 10-08-08. sigh.'verbose' in Class::AutoClass::Root conflicts with method in Base
 #              because AutoDB splices itself onto front of @ISA.
@@ -594,7 +831,7 @@ Data::Babel - Translator for biological identifiers
 
 =head1 VERSION
 
-Version 1.11
+Version 1.12
 
 =head1 SYNOPSIS
 
@@ -666,14 +903,14 @@ contain multiple Data::Babel objects; these may share some or all
 identifier types, and may provide the same or different mappings over
 the shared types.
 
-The principal method is 'translate' which converts identifiers of one
-type into identifiers of one or more output types.  In typical usage,
-you call 'translate' with a list of input ids to convert.  You can
-also call it without any input ids (or with the special option
+The principal method is L<"translate"> which converts identifiers of
+one type into identifiers of one or more output types.  In typical
+usage, you call L<"translate"> with a list of input ids to convert.
+You can also call it without any input ids (or with the special option
 'input_ids_all' set) to generate a complete mapping of the input type
 to the output types.  This is convenient if you want to hang onto the
 mapping for repeated use.  You can also filter the output based on
-values of other identifier types.  
+values of other identifier types.
 
 Comparisons are done in a B<case insensitive> manner.  This includes
 input ids, filters, and internal comparisons used to join database
@@ -722,9 +959,9 @@ histories). The query or program that loads the database is
 responsible for mapping old identifiers to current ones (presumably
 via the history).
 
-'translate' checks the input IdType to see if its Master has history
-information. If so, 'translate' automatically applies the history to
-all input ids.  It does the same for filters.
+L<"translate"> checks the input IdType to see if its Master has
+history information. If so, L<"translate"> automatically applies the
+history to all input ids.  It does the same for filters.
 
 You need not explicitly define Masters for all IdTypes; Babel
 will create 'implicit' Masters for any IdTypes lacking explicit
@@ -858,7 +1095,7 @@ MapTable
 
 This excerpt has two MapTable definitions which illustrate two ways
 that MapTables can be named.  The first uses a normal section name;
-the second invokes a L<Template Toolkit|Template> macro which generates unique
+the second invokes a L<Template::Toolkit> macro which generates unique
 names of the form 'maptable_001'.  This is very convenient because
 Babel databases typically contain a large number of MapTables, and
 it's hard to come up with good names for most of them.  In any case,
@@ -870,10 +1107,10 @@ construction procedure and may not be useful in other settings.
 
 =head2 Input ids that do not connect to any outputs
 
-By default, the 'translate' method does not return any output for
+By default, the L<"translate"> method does not return any output for
 input identifiers that do not connect to any identifiers of the
 desired output types; these are output rows in which the output
-columns are all NULL. You can instruct 'translate' to include these
+columns are all NULL. You can instruct L<"translate"> to include these
 rows in the result by setting the 'validate' option.
 
 An input identifier can fail to connect for two reasons: 
@@ -892,13 +1129,31 @@ If you set the 'validate' option, the output will contain at least one
 row for each input identifier, and an additional column that indicates
 whether each input identifier is valid.
 
-If no output IdTypes are specified, 'translate' returns a row
+If no output IdTypes are specified, L<"translate"> returns a row
 containing one element, namely, the input identifier, for each input
 id that exists in the corresponding Master table. If the 'validate'
 option is set, the output will contain one row for each input
 identifier; this is essentially a (possibly re-ordered) copy of the
 input list with duplicates removed.
 
+=head2 Partial duplicates
+
+A partial duplicate is a row that contains less information than
+another row and is therefore redundant.  More precisely, a row is a
+partial duplicate of another row if for all fields (1) the rows are
+identical or, (2) the field in the partial duplicate is NULL.  In the
+example below, the second row is a partial duplicate of the first.
+
+ gene_symbol  organism_name  gene_entrez  probe_id
+ HTT          human          3064         A_23_P212749
+ HTT          human          3064
+
+By default, L<"translate"> removes partial duplicates. The algorithm
+for removing partial duplicates may be slow for queries with a large
+number of output columns in cases where a given input id matches a
+large number of output ids. To retain partial duplicates, you can
+specify the 'keep_pdups' option to L<"translate">.
+
 =head2 Technical details
 
 A basic Babel property is that translations are stable. You can add
@@ -910,13 +1165,13 @@ everything that was in the original answer.
 
 We accomplish this by requiring that the database of MapTables satisfy
 the B<universal relation property> (a well-known concept in relational
-database theory), and that 'translate' retrieves a sub-table of the
+database theory), and that L<"translate"> retrieves a sub-table of the
 universal relational.  Concretely, the universal relational is the
-natural full outer join of all the MapTables.  'translate' performs
+natural full outer join of all the MapTables. L<"translate"> performs
 natural left out joins starting with the Master table for the input
 IdType, and then including enough tables to connect the input and
-output IdTypes. Left outer joins suffice, because 'translate' starts
-with the Master.
+output IdTypes. Left outer joins suffice, because L<"translate">
+starts with the Master.
 
 We further require that the database of MapTables be
 non-redundant. The basic idea is that a given IdType may not be
@@ -932,7 +1187,7 @@ undirected graph whose nodes represent IdTypes and MapTables, and
 whose edges go between each MapTable and the IdTypes it contains. In
 this representation, a non-redundant schema is a tree.
 
-'translate' uses this graph to find the MapTables it must join to
+L<"translate"> uses this graph to find the MapTables it must join to
 connect the input and output IdTypes. The algorithms is simple: start
 at the leaves and recursively prune back branches that do not contain
 the input or output IdTypes.
@@ -1036,6 +1291,8 @@ The available class attributes are
            count          boolean. If true, return number of output rows rather 
                           than the rows themselves. Equivalent to 'count'
                           method.
+           keep_pdups     boolean. If true, partial duplicates are not removed
+                          from the result.
 
 =head3 Notes on translate
 
@@ -1148,7 +1405,7 @@ only value for a filter type or whether it's one of several.
 
 =head3 Histories
 
-'translate' automatically applies histories, when they exist, to input
+L<"translate"> automatically applies histories, when they exist, to input
 and filter ids. In other words, input and filter ids can be ones that
 were valid in the past but are not valid now. Output ids, however, are
 always current.
@@ -1171,7 +1428,7 @@ can contain rows in which the output id does not match the filter.
  Returns : number
  Args    : same as 'translate'
 
-'count' is a wrapper for L<translate> that sets the 'count' argument to a true value.
+'count' is a wrapper for L<"translate"> that sets the 'count' argument to a true value.
 
 =head2 validate
 
@@ -1190,19 +1447,7 @@ can contain rows in which the output id does not match the filter.
                 be the same as the original id
            If output_idtypes is set, the result is ther same as 'translate' with
            the 'validate' option set
- Args    : input_idtype   name of Data::Babel::IdType object or object
-           input_ids      id or ARRAY of ids to be translated. If absent or
-                          undef, all ids of the input type are translated. If an
-                          empty array, ie, [], no ids are translated and the 
-                          result will be empty.
-           input_ids_all  boolean. If true, all ids of the input type are
-                          translated. Same as omitting input_ids or setting it
-                          to undef but more explicit.
-           output_idtypes optional and usually omitted. ARRAY of names of 
-                          Data::Babel::IdType objects or objects. If set,
-                          equivalent to calling 'translate' with the 'validate'
-                          option set
-           limit          maximum number of rows to retrieve
+ Args    : same as 'translate'
 
 'validate' looks up the given input ids in the Master tables for the
 given input type and returns a table indicating which ids are
@@ -1213,10 +1458,11 @@ the current value will always equal the given id if the id is valid.
 'validate' can also retrieve a complete table of valid ids (along with
 history information) for the type.
 
-'validate' is a wrapper for L<translate> that sets the 'validate'
-argument to a true value and the output_idtypes argument to the
-input_idtype.  All other 'translate' arguments (filters, count) are
-legal here and work but are of dubious value.
+'validate' is a wrapper for L<"translate"> that (1) sets the
+'validate' argument to a true value and (2) sets the output_idtypes
+argument to the input_idtype unless the user explicitly set it.  All
+other L<"translate"> arguments (filters, count) are legal here and
+work but are of dubious value.
 
 =head3 Notes on validate
 
